// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#include <fcntl.h>
#include <unistd.h>

#include <array>
#include <climits>
#include <cstdlib>
#include <cstring>
#include <filesystem>
#include <iostream>
#include <string>

#include "Common.hpp"
#include "Fiber.hpp"
#include "Runtime.hpp"
#include "Semaphore.hpp"
#include "io.hpp"

namespace fs = std::filesystem;

#define EMPER_RIPGREP_BUFSIZE 4096

const char* needle;
size_t needle_len;

emper::Semaphore* max_running;

void search(const std::string& path) {
	if (max_running) {
		max_running->acquire();
	}

	int fd = emper::io::openAndWait(path.c_str(), O_RDONLY);
	if (fd < 0) {
		DIE_MSG_ERRNO("open failed");
	}

	std::array<char, EMPER_RIPGREP_BUFSIZE> buf;
	off_t bytes_searched = 0;

	ssize_t bytes_read = emper::io::readFileAndWait(fd, buf.data(), buf.size(), bytes_searched);
	while (bytes_read > 0) {
		if (memmem(&buf[0], bytes_read, needle, needle_len)) {
			std::array<char, PATH_MAX + 1> outBuf;
			memcpy(outBuf.data(), path.c_str(), path.length());
			outBuf[path.length()] = '\n';
			emper::io::writeFileAndWait(STDOUT_FILENO, outBuf.data(), path.length() + 1, 0);
			goto out;
		}

		bytes_searched += bytes_read;
		bytes_read = emper::io::readFileAndWait(fd, buf.data(), buf.size(), -1);
	}

	if (bytes_read < 0) {
		DIE_MSG_ERRNO("read failed");
	}

out:
	if (max_running) {
		max_running->release();
	}

	emper::io::closeAndForget(fd);
}

void walk_dir() {
	// Only search in regular files
	auto filter = [](const fs::directory_entry& dirent) { return dirent.is_regular_file(); };
	// Search the file
	auto fn = [](const fs::directory_entry& dirent) { search(dirent.path()); };
	emper::io::recursive_directory_walk(".", filter, fn);

	Runtime::getRuntime()->initiateTermination();
}

auto main(int argc, char* argv[]) -> int {
	if (argc < 2) {
		std::cerr << "Usage: " << argv[0] << " <needle> [max fibers]" << std::endl;
		return EXIT_SUCCESS;
	}

	if (argc == 3) {
		max_running = new emper::Semaphore(std::stoi(argv[2]));
	}

	needle = argv[1];
	needle_len = strlen(needle);

	Runtime runtime;

	auto* dirWalker = Fiber::from(walk_dir);
	runtime.scheduleFromAnywhere(*dirWalker);

	runtime.waitUntilFinished();
	return EXIT_SUCCESS;
}
