// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus, Florian Fischer
#pragma once

#include <sstream>	// IWYU pragma: keep
#include <string>
#include <string_view>

#include "emper-config.h"

// If we apply clang-format to the following region, then clang-format
// will create multi-line macros. However clang-tidy's NOLINTNEXTLINE
// would then not work. :(

// clang-format off

// Hide the stringstream inside a lamba from the context where our LOG macros are used.
// The lamda should be fairly cheap/free because it will be inlined.
// The use of a lambda here is the ISO C++ equivalent to GCC statement expressions.
// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define EMPER_BUILD_STR(x) [&]() -> std::string {std::stringstream sst; sst << x; return sst.str(); }()

// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define LOG(level, x, log_func) do { if constexpr (level <= EMPER_LOG_LEVEL) { log_func(EMPER_BUILD_STR(x)); } } while (false)

// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define DBG(x) LOG(Debug, x, emper_log_no_prefix);

// To avoid "error: there are no arguments to ‘logD’ that depend on a
// template parameter, so a declaration of ‘logD’ must be available"
// we use "this->logD()" instead of simply "logD()" below.
// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define LOGD(x) LOG(Debug, x, this->logD);
// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define LOGDD(x) LOG(FineDebug, x, this->logDD);

// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define LOGI(x) LOG(Info, "Info: " << x, emper_log_no_prefix);
// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define LOGGER_LOGI(x) LOG(Info, "Info: " << x, this->logI);

// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define LOGW(x) LOG(Warning, "Warning: " << x, emper_log_no_prefix);
// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define LOGGER_LOGW(x) LOG(Warning, "Warning: " << x, this->logW);

// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define LOGE(x) LOG(Error, "Error: " << x, emper_log_no_prefix);
// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define LOGGER_LOGE(x) LOG(Error, "Error: " << x, emper_log_no_prefix);

// NOLINTNEXTLINE(bugprone-macro-parentheses)
#define ABORT(x) { LOGE(x); abort(); }

// clang-format on

#define IGNORE_UNUSED_FUNCTION \
	_Pragma("GCC diagnostic push") _Pragma("GCC diagnostic ignored \"-Wunused-function\"")

// test for GCC > 11.1.0
#if __GNUC__ > 11 || \
		(__GNUC__ == 11 && (__GNUC_MINOR__ > 1 || (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 0)))
#define IGNORE_MAYBE_UNINITIALIZED \
	_Pragma("GCC diagnostic push") _Pragma("GCC diagnostic ignored \"-Wmaybe-uninitialized\"")
#else
#define IGNORE_MAYBE_UNINITIALIZED _Pragma("GCC diagnostic push")
#endif

#define POP_DIAGNOSTIC _Pragma("GCC diagnostic pop")

enum class LogSubsystem {
	PS,
	F,
	C,
	CM,
	DISP,
	SCHED,
	RUNTI,
	SLEEP_S,
	U_B_MPSC_Q,
	IO,
};

enum LogLevel {
	Error,
	Warning,
	Info,
	Debug,
	FineDebug,
	FinerDebug,
	FinestDebug,
	ALL,
};

void emper_add_timestamp_to(std::ostringstream& logMessage);

void emper_log(const std::string& prefix, const std::string& message);

static inline void emper_log_no_prefix(const std::string& message) { emper_log("", message); }

template <LogSubsystem logSubsystem>
class Logger {
 private:
	static constexpr auto getLevelFor(LogSubsystem system) -> LogLevel {
		switch (system) {
			case LogSubsystem::PS:
			case LogSubsystem::F:
			case LogSubsystem::C:
			case LogSubsystem::CM:
			case LogSubsystem::DISP:
			case LogSubsystem::SCHED:
			case LogSubsystem::RUNTI:
			case LogSubsystem::SLEEP_S:
			case LogSubsystem::U_B_MPSC_Q:
			case LogSubsystem::IO:
			default:
				return ALL;
		}
	}

	static constexpr auto getTagFor(LogSubsystem system) -> std::string_view {
		switch (system) {
			case LogSubsystem::PS:
				return "PS   ";
			case LogSubsystem::F:
				return "F    ";
			case LogSubsystem::C:
				return "C    ";
			case LogSubsystem::CM:
				return "CM   ";
			case LogSubsystem::DISP:
				return "DISP ";
			case LogSubsystem::SCHED:
				return "SCHED";
			case LogSubsystem::RUNTI:
				return "RUNTI";
			case LogSubsystem::SLEEP_S:
				return "SLEEP_S";
			case LogSubsystem::U_B_MPSC_Q:
				return "UBSCQ";
			case LogSubsystem::IO:
				return "IO   ";
			default:
				return "UNKNOWN SUBSYSTEM (Add it *now*)";
		}
	}

	static constexpr auto shouldPrefixThis(LogSubsystem system) -> bool {
		switch (system) {
			case LogSubsystem::RUNTI:
			case LogSubsystem::SCHED:
				return false;
			default:
				return true;
		}
	}

 protected:
	template <enum LogLevel level>
	inline void log(const std::string& string) const {
		// check global level
		if constexpr (level > EMPER_LOG_LEVEL) return;
		// check subsystem level
		if constexpr (level > getLevelFor(logSubsystem)) return;

		constexpr std::string_view subSystemTag = getTagFor(logSubsystem);

		std::ostringstream sst;
		sst << subSystemTag;
		if constexpr (shouldPrefixThis(logSubsystem)) {
			sst << " " << this;
		}

		emper_log(sst.str(), string);
	}

	inline void logE(const std::string& string) const { log<Error>(string); }

	inline void logI(const std::string& string) const { log<Info>(string); }

	inline void logD(const std::string& string) const { log<Debug>(string); }

	inline void logDD(const std::string& string) const { log<FineDebug>(string); }
};
