// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus
#pragma once

#include "Common.hpp"
#include "Runtime.hpp"
#include "emper-common.h"
#include "lib/adt/BoundedBumpArray.hpp"
#include "lib/adt/WsClQueue.hpp"

template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
class MemoryManager {
 private:
	const workerid_t workerCount;

	adt::WsClQueue<void*, WS_QUEUE_SIZE>** queues;

	static thread_local adt::BoundedBumpArray<void, WORKER_EXCLUSIVE_QUEUE_SIZE> workerExclusiveQueue;

	static thread_local adt::WsClQueue<void*, WS_QUEUE_SIZE> queue;

	static auto mallocMemory() -> void* {
		void* memory;
		int res = posix_memalign(&memory, alignof(T), sizeof(T));
		if (res != 0) {
			abort();
		}
		return memory;
	}

 public:
	MemoryManager(Runtime& runtime);

	auto getMemory(bool* malloced) -> void* {
		*malloced = false;
		void* memory = workerExclusiveQueue.get();
		if (memory) return memory;

		bool poped = queue.popTop(&memory);

		if (likely(poped)) return memory;

		// Fallback to memory-stealing.
		const workerid_t myWorkerId = Runtime::getWorkerId();
		const workerid_t startWorkerId = Runtime::rand() % workerCount;
		for (workerid_t i = 0; i < workerCount; ++i) {
			workerid_t victim = (startWorkerId + i) % workerCount;

			// Don't steal from ourselves.
			if (unlikely(victim == myWorkerId)) continue;

			poped = queues[victim]->popTop(&memory);
			if (poped) return memory;
		}

		*malloced = true;
		// If everything fails, allocate the memory.
		return mallocMemory();
	}

	void putMemory(T* memory) {
		assert(memory);
		bool pushed = workerExclusiveQueue.put(memory);
		if (pushed) return;

		pushed = queue.pushBottom(memory);
		if (pushed) return;

		free(memory);
	}
};

template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
thread_local adt::BoundedBumpArray<void, WORKER_EXCLUSIVE_QUEUE_SIZE>
		MemoryManager<T, WS_QUEUE_SIZE, WORKER_EXCLUSIVE_QUEUE_SIZE>::workerExclusiveQueue;

template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
thread_local adt::WsClQueue<void*, WS_QUEUE_SIZE>
		MemoryManager<T, WS_QUEUE_SIZE, WORKER_EXCLUSIVE_QUEUE_SIZE>::queue;

template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
MemoryManager<T, WS_QUEUE_SIZE, WORKER_EXCLUSIVE_QUEUE_SIZE>::MemoryManager(Runtime& runtime)
		: workerCount(runtime.getWorkerCount()) {
	queues = new adt::WsClQueue<void*, WS_QUEUE_SIZE>*[workerCount];

	auto newWorkerHook = [this](workerid_t workerId) { queues[workerId] = &queue; };
	runtime.addNewWorkerHook(newWorkerHook);
}
