// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus, Florian Fischer
#include "Runtime.hpp"

#include <pthread.h>	// for pthread_t, pthread_attr_init

#include <cerrno>	 // for errno
// Non portable.
#include <sched.h>				// for cpu_set_t, CPU_SET, CPU_ZERO
#include <sys/sysinfo.h>	// for get_nprocs

#include <cstdlib>	// for rand, srand, abort
#include <cstring>
#include <iostream>
#include <memory>	 // for __shared_ptr_access, shared_ptr
#include <string>	 // for string
#include <thread>

#include "Common.hpp"	 // for DIE_MSG_ERRNO, DIE, DIE_MSG
#include "Context.hpp"
#include "ContextManager.hpp"	 // for ContextManager
#include "Debug.hpp"					 // for DBG, ABORT, LOGD, LOGE
#include "Emper.hpp"
#include "Fiber.hpp"	// for Fiber
#include "NextFiberResult.hpp"
#include "RuntimeStrategy.hpp"	// for RuntimeStrategy
#include "RuntimeStrategyFactory.hpp"
#include "RuntimeStrategyStats.hpp"	 // for RuntimeStrategyStats
#include "WorkerLocalData.hpp"
#include "emper-config.h"
#include "io/GlobalIoContext.hpp"
#include "io/IoContext.hpp"
#include "io/Stats.hpp"
#include "lib/DebugUtil.hpp"
#include "stats/FromAnywhere.hpp"
#include "stats/Worker.hpp"
#include "strategies/AbstractWorkStealingScheduler.hpp"

#ifdef EMPER_DEFAULT_SCHEDULING_STRATEGY_WORK_STEALING
#include "strategies/ws/WsStrategyFactory.hpp"
#elif defined EMPER_DEFAULT_SCHEDULING_STRATEGY_LOCALITY_AWARE_WORK_STEALING
#include "strategies/laws/LawsStrategyFactory.hpp"
#else
#error "Unknown default scheduling strategy"
#endif

#ifdef EMPER_LIBURCU
#include <urcu.h>	 // for rcu_register_thread
#endif

#include <syscall.h>	// for SYS_gettid
#include <unistd.h>		// for syscall

std::mutex Runtime::currentRuntimeMutex;
Runtime* Runtime::currentRuntime;

RuntimeStrategyFactory& Runtime::DEFAULT_STRATEGY =
#ifdef EMPER_DEFAULT_SCHEDULING_STRATEGY_WORK_STEALING
		WsStrategyFactory::INSTANCE
#elif defined EMPER_DEFAULT_SCHEDULING_STRATEGY_LOCALITY_AWARE_WORK_STEALING
		LawsStrategyFactory::INSTANCE
#else
#error "Unknown default scheduling strategy"
#endif
		;

using emper::io::GlobalIoContext;
using emper::io::IoContext;

Runtime::Runtime(workerid_t workerCount, RuntimeStrategyFactory& strategyFactory, unsigned int seed)
		: workerCount(workerCount),
			workerLatch(workerCount),
			firstWorkerThreadExitLatch(workerCount),
			secondWorkerThreadExitLatch(workerCount),
			strategy(strategyFactory.constructRuntimeStrategy(*this)),
			scheduler(strategy->getScheduler()),
			dispatcher(strategy->getDispatcher()),
			contextManager(*(new ContextManager(*this))),
			threads(new pthread_t[workerCount]),
			workers(new Worker*[workerCount]),
			ioContexts(emper::IO ? workerCount : 0),
			ioReadySem(0),
			randomEngine(seed),
			workerSleepStrategy(workerCount) {
	const int nprocs = get_nprocs();

	{
		std::lock_guard<std::mutex> lock(currentRuntimeMutex);
		if (currentRuntime != nullptr) {
			DIE_MSG("There is already a runtime active");
		}
		currentRuntime = this;
	}

	if constexpr (emper::STATS) {
		workerStats = new WorkerLocalData<emper::stats::Worker>(
				[workerCount] {
					emper::stats::worker = new emper::stats::Worker(workerCount);
					return emper::stats::worker;
				},
				*this);
		fromAnywhereStats = new emper::stats::FromAnywhere(workerCount);
	} else {
		workerStats = nullptr;
		fromAnywhereStats = nullptr;
	}

	// initialize the global IoContext if a completer is used
	if constexpr (emper::IO && emper::IO_COMPLETER_BEHAVIOR != emper::IoCompleterBehavior::none) {
		// The global io_uring needs at least workerCount entries in its SQ because
		// for each worker's IoContext one eventfd read is prepared before the
		// globalCompleter is started and submits all previously prepared sqes.
		globalIo = new GlobalIoContext(*this, workerCount);
		globalIo->startGlobalCompleter();

		if constexpr (emper::STATS) {
			globalIo->stats.workerId = emper::io::Stats::GLOBAL_COMPLETER_ID;
		}
	}

	bool pinWorkers = shouldPinWorkers();

	// Core id we start the worker pinning
	workerid_t pinningOffset = 0;
	char* pinningOffsetEnv = std::getenv("EMPER_PINNING_OFFSET");
	if (pinningOffsetEnv) {
		if (!pinWorkers) {
			DIE_MSG("EMPER_PIN_WORKERS=false and EMPER_PINNING_OFFSET are mutual exclusive");
		}

		int pinningOffsetInt = std::stoi(pinningOffsetEnv);
		if (pinningOffsetInt > UINT8_MAX) {
			DIE_MSG("Pinning offset " << pinningOffsetInt << " to big for its datatype");
		}
		pinningOffset = static_cast<workerid_t>(pinningOffsetInt);
	}

	for (workerid_t i = 0; i < workerCount; ++i) {
		pthread_attr_t attr;
		errno = pthread_attr_init(&attr);
		if (errno) DIE;

		if (pinWorkers) {
			// Start non portable.
			cpu_set_t cpuset;
			CPU_ZERO(&cpuset);
			CPU_SET((i + pinningOffset) % nprocs, &cpuset);

			errno = pthread_attr_setaffinity_np(&attr, sizeof(cpuset), &cpuset);
			if (errno) DIE_MSG_ERRNO("pthread_attr_setaffinity_np() failed");
			// End non portable.
		}

		// This is not done by each thread individually because the randomEngine is
		// mutated and therefore produces data races.
		unsigned int seed = uniformIntDistribution(randomEngine);
		auto* worker = new Worker(i, workerCount, seed);
		workers[i] = worker;

		auto thread_function = [](void* voidWorkerPtr) -> void* {
#ifdef EMPER_LIBURCU
			rcu_register_thread();
#endif

			auto* worker = reinterpret_cast<Worker*>(voidWorkerPtr);
			return currentRuntime->workerLoop(worker);
		};
		errno = pthread_create(&threads[i], &attr, thread_function, worker);
		if (errno) DIE_MSG_ERRNO("pthread_create() failed");
	}

	threadsRunning = true;

	const std::string stacktraceOnAbortsEnvVar("EMPER_STACKTRACE_ON_ABORTS");
	const char* envValueEnableStacktraces = std::getenv(stacktraceOnAbortsEnvVar.c_str());
	if (envValueEnableStacktraces && !std::strcmp(envValueEnableStacktraces, "true")) {
		DBG(stacktraceOnAbortsEnvVar << " set to 'true': Enabling stacktrace on abort");
		enableStacktraceOnAborts();
	}
}

Runtime::~Runtime() {
	DBG("Runtime " << this << " is terminating");
	if (threadsRunning) {
		initiateAndWaitUntilTermination();
	}

	// Print the stats after all worker threads have terminated and before we delete
	// objects bound to the runtime's lifetime
	if constexpr (emper::STATS) {
		printLastRuntimeStats();

		// TODO: The following should probably be part of printLastRuntimeStats?
		if constexpr (emper::IO) {
			emper::io::Stats::printStats(globalIo, ioContexts);
		}

		workerSleepStrategy.printStats();
	}

	for (unsigned int i = 0; i < workerCount; ++i) {
		delete workers[i];
	}
	delete[] workers;
	delete[] threads;

	if constexpr (emper::IO) {
		if constexpr (emper::IO_COMPLETER_BEHAVIOR != emper::IoCompleterBehavior::none) {
			// It is safer to destroy the globalIo before the worker IoContexts
			// because pointer to the worker IoContexts can outlife their objects
			// in the globalCompleter
			delete globalIo;
		}

		for (unsigned int i = 0; i < workerCount; ++i) {
			delete ioContexts[i];
		}
	}

	if constexpr (emper::STATS) {
		workerStats->forEach([](auto* stats) { delete stats; });
		delete workerStats;
		delete fromAnywhereStats;
	}

	// TODO: Move this before the delete invocations?
	{
		std::lock_guard<std::mutex> lock(currentRuntimeMutex);
		currentRuntime = nullptr;
	}
	DBG("Runtime " << this << " terminated");
}

auto Runtime::workerLoop(Worker* worker) -> void* {
	worker->setWorker();

	if constexpr (emper::IO) {
		auto* workerIo = new IoContext(*this);

		if constexpr (emper::IO_COMPLETER_BEHAVIOR != emper::IoCompleterBehavior::none) {
			// submit the workers' CQ eventfds to the global IoContext
			globalIo->registerWorkerIo(*workerIo);
		}
		// notify the globalCompleter that we have registered our eventfd
		ioReadySem.notify();

		ioContexts[worker->workerId] = workerIo;
		workerIo->setWorkerIo(worker);
	}

	LOGD("Worker loop started by thread " << syscall(SYS_gettid));

	int oldType;
	errno = pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &oldType);
	if (errno) {
		DIE_MSG_ERRNO("pthread_setcanceltype() failed");
	}

	for (const auto& f : newWorkerHooks) f(worker->workerId);

	workerLatch.count_down_and_wait();

	contextManager.start();

	// Threads return here if Context::switchToOriginalStack() is called.

	// Ensure that all worker threads exit "at the same" time. Otherwise
	// it would be possible that one thread is work-stealing,
	// potentially accessing a work stealing queue of an worker thread
	// that already exited, causing an invalid access.
	firstWorkerThreadExitLatch.count_down_and_wait();

	if (worker->getWorkerId() == 0) {
		// Obtain stats when all workers reached the first latch but before they exit.
		strategyLastStats = strategy->getStats();
		// Note that we do not need to do the same for the
		// RuntimeWorkerStats, as these are dynamically allocated and
		// still available once the worker threads terminate.
	}

	secondWorkerThreadExitLatch.count_down_and_wait();

	return nullptr;
}

auto Runtime::getDefaultWorkerCount() -> workerid_t {
	char* workerCountEnv = std::getenv("EMPER_WORKER_COUNT");
	if (workerCountEnv) {
		int workerCountInt = std::stoi(workerCountEnv);
		if (workerCountInt > UINT8_MAX) {
			DIE_MSG("Worker count " << workerCountInt << " to big for its datatype");
		}
		return static_cast<workerid_t>(workerCountInt);
	}

	// The CPU count reported by sysconf(_SC_NPROCESSORS_ONLN), sysconf(_SC_NPROCESSORS_CONF)
	// or std::thread::hardware_concurrency() may not match the CPU count available
	// to the emper process.
	// This can happen for example if we run EMPER in a container or qemu.o
	// GNU nproc(1) does report the correct CPU count using pthread_getaffinity_np
	// which fills a cpu_set with the CPUs the current process can be scheduled on
	// therefore we use this more precise CPU count if possible.
	// Code inspiration taken from:
	// https://github.com/coreutils/gnulib/blob/90e79512d8b385801218d6e9c4d88ff77186560b/lib/nproc.c#L206
	cpu_set_t set;
	if (pthread_getaffinity_np(pthread_self(), sizeof(set), &set) == 0) {
		unsigned long count = CPU_COUNT(&set);
		if (count > 0) {
			return count;
		}
	}

	return std::thread::hardware_concurrency();
}

void Runtime::yield() {
	Context* context = Context::getCurrentContext();
	contextManager.saveAndStartNew([context, this] {
		Fiber* resumeFiber =
				Fiber::from([context, this] { this->contextManager.discardAndResume(context); });
		this->scheduleFromAnywhere(*resumeFiber);
	});
}

auto Runtime::nextFiber() -> NextFiberResult {
	if constexpr (emper::IO) {
		// Schedule all fibers waiting on completed IO
		IoContext::ContinuationBuffer completions;
		unsigned ncompletions = IoContext::getWorkerIo()->reapCompletions(completions);
		if (ncompletions > 0) {
			// Keep the first and schedule the rest
			Fiber* next = completions[0];
			schedule(&completions[1], ncompletions - 1);

			// TODO: hint that this fiber comes from the IO subsystem
			return NextFiberResult{
					next, static_cast<uintptr_t>(AbstractWorkStealingScheduler::FiberSource::local)};
		}
	}

	return scheduler.nextFiber();
}

void Runtime::initiateTermination() {
	terminateWorkers.store(true, std::memory_order_release);

	workerSleepStrategy.notifyAll<CallerEnvironment::ANYWHERE>();

	if constexpr (emper::IO && emper::IO_COMPLETER_BEHAVIOR != emper::IoCompleterBehavior::none) {
		globalIo->initiateTermination();
	}
}

void Runtime::waitUntilFinished() {
	for (workerid_t i = 0; i < workerCount; ++i) {
		errno = pthread_join(threads[i], nullptr);
		if (errno) {
			DIE_MSG_ERRNO("pthread_join() returned error for worker ID " << std::to_string(i));
		}
	}

	if constexpr (emper::IO && emper::IO_COMPLETER_BEHAVIOR != emper::IoCompleterBehavior::none) {
		globalIo->waitUntilFinished();
	}

	threadsRunning = false;
}

void Runtime::initiateAndWaitUntilTermination() {
	initiateTermination();
	waitUntilFinished();
}

void Runtime::printStats(bool detailed) {
	if constexpr (!emper::STATS) {
		std::cout << "printStats(): EMPER stats disabled by static configuration" << std::endl;
		return;
	}

	std::shared_ptr<RuntimeStrategyStats> stats;
	if (terminateWorkers) {
		// The runtime was already requested to shut down. We can not
		// obtain stats as it is not guranteed that the worker threads are
		// still running in this case. We can however check
		// strategyLastStats and print that.

		if (strategyLastStats) {
			stats = strategyLastStats;
		} else {
			std::cerr << "printStats(): worker threads have already been scheduled for termination, but "
									 "not last stats recored yet."
								<< std::endl;
			return;
		}
	} else {
		stats = strategy->getStats();
	}

	stats->print();

	emper::stats::Worker globalWorkerStats(workerCount);
	auto workerStatsSnapshot = workerStats->getSnapshot();
	for (auto& stats : workerStatsSnapshot) {
		globalWorkerStats += stats;
	}

	globalWorkerStats.print();
	if (detailed) {
		for (auto& stats : workerStatsSnapshot) {
			stats.print();
		}
	}

	fromAnywhereStats->print();
}

void Runtime::printLastRuntimeStats() {
	std::lock_guard<std::mutex> lock(currentRuntimeMutex);
	if (currentRuntime == nullptr) {
		LOGE("Could not print current runtime stats");
		return;
	}

	currentRuntime->printStats();
}

void Runtime::executeAndWait(std::function<void()> f) {
	if (inRuntime()) {
		ABORT("Ca not use executeAndWait() from within the Runtime");
	}

	std::mutex fiberFinished;
	fiberFinished.lock();

	Fiber* fiber = Fiber::from([&] {
		f();

		fiberFinished.unlock();
	});

	scheduleFromAnywhere(*fiber);

	fiberFinished.lock();
}
