// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#pragma once

#include <fcntl.h>
#include <sys/socket.h>	 // for socklen_t
#include <sys/types.h>	 // for ssize_t, off_t

#include <cstddef>	// for size_t
#include <filesystem>
#include <functional>	 // for function
#include <memory>			 // for unique_ptr
#include <string>			 // for string

namespace fs = std::filesystem;

#include "Future.hpp"	 // for Futures
#include "emper-config.h"

/*
 * Header defining the public POSIX-like IO interface of emper.
 * All functions defined in this header must be used inside a fiber.
 * The emper IoContext to handle the IO request will be the one setup by the
 * Runtime and returned by IoContext::getWorkerIo().
 */

class Fiber;

namespace emper::io {
/**
 * @brief non-blocking recv mimicking POSIX recv(3)
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param socket file descriptor for this operation
 * @param buffer destination buffer
 * @param length length of the message in bytes
 * @param flags type of message reception
 *
 * @return Future object which signals the completion of the recv request
 */
[[nodiscard]] inline auto recv(int socket, void *buffer, size_t length, int flags)
		-> std::unique_ptr<Future> {
	auto future = std::make_unique<RecvFuture>(socket, buffer, length, flags);
	future->submit();
	return future;
}

/**
 * @brief Blocking recv mimicking POSIX recv(3)
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param socket file descriptor for this operation
 * @param buffer destination buffer
 * @param length length of the message in bytes
 * @param flags type of message reception

 * @return -1 on error, 0 when receiving from a closed socket, otherwise the received bytes
 */
inline auto recvAndWait(int socket, void *buffer, size_t length, int flags) -> ssize_t {
	RecvFuture future(socket, buffer, length, flags);
	future.submit();
	return future.waitAndSetErrno();
}

/**
 * @brief Blocking recv mimicking POSIX recv(3) with timeout
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param socket file descriptor for this operation
 * @param buffer destination buffer
 * @param length length of the message in bytes
 * @param flags type of message reception
 * @param ts Time to wait before aborting the request

 * @return -1 on error, 0 when receiving from a closed socket, otherwise the received bytes
 */
inline auto recvAndTryWait(int socket, void *buffer, size_t length, int flags,
													 TimeoutWrapper::Timespec &ts) -> ssize_t {
	RecvFuture future(socket, buffer, length, flags);
	TimeoutWrapper wrapper(future, ts);
	wrapper.submitAndWait();
	return future.waitAndSetErrno();
}

/**
 * @brief Non-blocking send mimicking POSIX send(3)
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param socket file descriptor for this operation
 * @param buffer source buffer
 * @param length length of the message in bytes
 * @param flags type of message transmission
 * @param send_all If true sginal the future only if all bytes are sent or
 *        sending further bytes is not possible
 *
 * @return Future object which signals the completion of the send request
 */
[[nodiscard]] inline auto send(int socket, const void *buffer, size_t length, int flags,
															 bool send_all = true) -> std::unique_ptr<Future> {
	auto future = std::make_unique<SendFuture>(socket, buffer, length, flags, send_all);
	future->submit();
	return future;
}

/**
 * @brief Blocking send mimicking POSIX send(3)
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param socket file descriptor for this operation
 * @param buffer source buffer
 * @param length length of the message in bytes
 * @param flags type of message transmission
 * @param send_all If true return to the user only if all bytes are sent or
 *        sending further bytes is not possible
 *
 * @return -1 on error, otherwise the number of sent bytes
 */
inline auto sendAndWait(int socket, const void *buffer, size_t length, int flags,
												bool send_all = true) -> ssize_t {
	void *mut_buf = const_cast<void *>(buffer);
	SendFuture future(socket, mut_buf, length, flags, send_all);
	future.submit();
	return future.waitAndSetErrno();
}

/**
 * @brief Blocking send mimicking POSIX send(3) with timeout
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param socket file descriptor for this operation
 * @param buffer source buffer
 * @param length length of the message in bytes
 * @param flags type of message transmission
 * @param ts Time to wait before aborting the request
 * @param send_all If true return to the user only if all bytes are sent or
 *        sending further bytes is not possible
 *
 * @return -1 on error, otherwise the number of sent bytes
 */
inline auto sendAndTryWait(int socket, const void *buffer, size_t length, int flags,
													 TimeoutWrapper::Timespec &ts, bool send_all = true) -> ssize_t {
	void *mut_buf = const_cast<void *>(buffer);
	SendFuture future(socket, mut_buf, length, flags, send_all);
	TimeoutWrapper wrapper(future, ts);
	wrapper.submitAndWait();
	return future.waitAndSetErrno();
}

/**
 * @brief Non-blocking connect mimicking POSIX connect(3)
 *
 * It is discouraged to call connect on a socket with the SOCK_NONBLOCK flag set
 * because the request submitted to io_uring will immediately be completed
 * with EINPROGESS or EALREADY and must be repeatedly reinserted into the submission
 * queue and io_uring afterwards.
 * To efficiently connect a socket don't set its SOCK_NONBLOCK flag.
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param socket file descriptor for this operation
 * @param address buffer containing the address to connect to
 * @param address_len length of the sockaddr structure
 *
 * @return Future object which signals the completion of the connect request
 */
[[nodiscard]] inline auto connect(int socket, const struct sockaddr *address, socklen_t address_len)
		-> std::unique_ptr<Future> {
	auto future = std::make_unique<ConnectFuture>(socket, address, address_len);
	future->submit();
	return future;
}

/**
 * @brief Blocking connect mimicking POSIX connect(3)
 *
 * It is discouraged to call connect on a socket with the SOCK_NONBLOCK flag set
 * because the request submitted to io_uring will immediately be completed
 * with EINPROGESS or EALREADY and must be repeatedly reinserted into the submission
 * queue and io_uring afterwards.
 * To efficiently connect a socket don't set its SOCK_NONBLOCK flag.
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param socket file descriptor for this operation
 * @param address buffer containing the address to connect to
 * @param address_len length of the sockaddr structure
 *
 * @return -1 on error, 0 otherwise
 */
inline auto connectAndWait(int socket, const struct sockaddr *address, socklen_t address_len)
		-> int {
	ConnectFuture future(socket, address, address_len);
	future.submit();
	return future.waitAndSetErrno();
}

/**
 * @brief Non-blocking accept mimicking POSIX accept(3)
 *
 * It is discouraged to call accept on a socket with the SOCK_NONBLOCK flag set
 * because the request submitted to io_uring will immediately be completed
 * with EAGAIN or EWOULDBLOCK and must be repeatedly reinserted into the submission
 * queue and io_uring afterwards.
 * To efficiently accept new connections don't set the SOCK_NONBLOCK flag.
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param socket file descriptor for this operation
 * @param[out] address buffer to store the clients address
 * @param[in, out] address_len length of the supplied sockaddr structure
 *
 * @return Future object which signals the completion of the accept request
 */
[[nodiscard]] inline auto accept(int socket, struct sockaddr *address, socklen_t *address_len)
		-> std::unique_ptr<Future> {
	auto future = std::make_unique<AcceptFuture>(socket, address, address_len);
	future->submit();
	return future;
}

/**
 * @brief Blocking accept mimicking POSIX accept(3)
 *
 * It is discouraged to call accept on a socket with the SOCK_NONBLOCK flag set
 * because the request submitted to io_uring will immediately be completed
 * with EAGAIN or EWOULDBLOCK and must be repeatedly reinserted into the submission
 * queue and io_uring afterwards.
 * To efficiently accept new connections don't set the SOCK_NONBLOCK flag.
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param socket file descriptor for this operation
 * @param[out] address buffer to store the clients address
 * @param[in, out] address_len length of the supplied sockaddr structure
 *
 * @return -1 on error, otherwise the non-negative file descriptor of the accepted socket.
 */
inline auto acceptAndWait(int socket, struct sockaddr *address, socklen_t *address_len) -> int {
	AcceptFuture future(socket, address, address_len);
	future.submit();
	return future.waitAndSetErrno();
}

/**
 * @brief Non-blocking read for regular files mimicking POSIX read(3)
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param fildes file descriptor to the regular file to be read from
 * @param buf destination buffer
 * @param nbyte amount of bytes to read
 * @param offset offset in the file
 * @param read_all If true signal the future  only if the buffer is completely
 *        filled or no further data can be read.
 *
 * @return Future object which signals the completion of the read request
 */
[[nodiscard]] inline auto readFile(int fildes, void *buf, size_t nbyte, off_t offset = -1,
																	 bool read_all = false) -> std::unique_ptr<Future> {
	auto future = std::make_unique<ReadFuture>(fildes, buf, nbyte, offset, read_all);
	future->submit();
	return future;
}

/**
 * @brief Blocking read for regular files mimicking POSIX read(3)
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param fildes file descriptor to the regular file to be written to
 * @param buf destination buffer
 * @param nbyte amount of bytes to read
 * @param offset offset in the file
 * @param read_all If true return only to the user if the buffer is completely
 *        filled or no further data can be read.
 *
 * @return -1 on error, otherwise the number of bytes read
 */
inline auto readFileAndWait(int fildes, void *buf, size_t nbyte, off_t offset = -1,
														bool read_all = false) -> ssize_t {
	ReadFuture future(fildes, buf, nbyte, offset, read_all);
	future.submit();
	return future.waitAndSetErrno();
}

/**
 * @brief Blocking read for regular files mimicking POSIX recv(3) with timeout
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param fildes file descriptor to the regular file to be written to
 * @param buf destination buffer
 * @param nbyte amount of bytes to read
 * @param offset offset in the file
 * @param ts Time to wait before aborting the request
 * @param read_all If true return only to the user if the buffer is completely
 *        filled or no further data can be read.
 *
 * @return -1 on error, 0 when receiving from a closed socket, otherwise the received bytes
 */
inline auto readFileAndTryWait(int fildes, void *buf, size_t nbyte, TimeoutWrapper::Timespec &ts,
															 off_t offset = -1, bool read_all = false) -> ssize_t {
	ReadFuture future(fildes, buf, nbyte, offset, read_all);
	TimeoutWrapper wrapper(future, ts);
	wrapper.submitAndWait();
	return future.waitAndSetErrno();
}

/**
 * @brief Non-blocking write for regular files mimicking POSIX write(3)
 *
 * Currently only reading from regular files is tested and supported.
 * Reading from other file types could work but may result in undefined behavior.
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param fildes file descriptor to the regular file to be written to
 * @param buf source buffer
 * @param nbyte amount of bytes to write
 * @param offset offset in the file
 * @param write_all If true signal this future only all nbyte were written
 *        or no further data can be written.
 *
 * @return Future object which signals the completion of the write request
 */
[[nodiscard]] inline auto writeFile(int fildes, const void *buf, size_t nbyte, off_t offset = -1,
																		bool write_all = true) -> std::unique_ptr<Future> {
	auto future = std::make_unique<WriteFuture>(fildes, buf, nbyte, offset, write_all);
	future->submit();
	return future;
}

/**
 * @brief Blocking write for regular files mimicking POSIX write(3)
 *
 * Currently only writing to regular files is tested and supported.
 * Writing to other file types could work but may result in undefined behavior.
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param fildes file descriptor to the regular file to be written to
 * @param buf source buffer
 * @param nbyte amount of bytes to write
 * @param offset offset in the file
 * @param write_all If true return only to the user if all nbyte were written
 *        or no further data can be written.
 *
 * @return -1 on error, otherwise the number of bytes written
 */
inline auto writeFileAndWait(int fildes, const void *buf, size_t nbyte, off_t offset = -1,
														 bool write_all = true) -> ssize_t {
	WriteFuture future(fildes, buf, nbyte, offset, write_all);
	future.submit();
	return future.waitAndSetErrno();
}

/**
 * @brief Blocking write for regular files mimicking POSIX write(3) with timeout
 *
 * Currently only writing to regular files is tested and supported.
 * Writing to other file types could work but may result in undefined behavior.
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param fildes file descriptor to the regular file to be written to
 * @param buf source buffer
 * @param nbyte amount of bytes to write
 * @param ts Time to wait before aborting the request
 * @param offset offset in the file
 * @param write_all If true return only to the user if all nbyte were written
 *        or no further data can be written.
 *
 * @return -1 on error, otherwise the number of bytes written
 */
inline auto writeFileAndTryWait(int fildes, const void *buf, size_t nbyte,
																TimeoutWrapper::Timespec &ts, off_t offset = -1,
																bool write_all = true) -> ssize_t {
	WriteFuture future(fildes, buf, nbyte, offset, write_all);
	TimeoutWrapper wrapper(future, ts);
	wrapper.submitAndWait();
	return future.waitAndSetErrno();
}

/**
 * @brief Non-blocking writev mimicking POSIX writev(3)
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param fildes file descriptor to write to
 * @param iov pointer to the iovec struct holding buffers to write
 * @param iovcnt number of buffers in the iovec array
 *
 * @return Future object which signals the completion of the write request
 */
[[nodiscard]] inline auto writev(int fildes, const struct iovec *iov, int iovcnt)
		-> std::unique_ptr<Future> {
	auto future = std::make_unique<WritevFuture>(fildes, iov, iovcnt);
	future->submit();
	return future;
}

/**
 * @brief Blocking writev mimicking POSIX writev(3)
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param fildes file descriptor to write to
 * @param iov pointer to the iovec struct holding buffers to write
 * @param iovcnt number of buffers in the iovec array
 *
 * @return -1 on error, otherwise the number of bytes written
 */
inline auto writevAndWait(int fildes, const struct iovec *iov, int iovcnt) -> ssize_t {
	WritevFuture future(fildes, iov, iovcnt);
	future.submit();
	return future.waitAndSetErrno();
}

/**
 * @brief asynchronous openat mimicking POSIX openat(3)
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param dirfd directory the pathname is interpreted relative to
 * @param pathname path to the file to open
 * @param flags specify access mode
 * @param mode in which the file should be created
 *
 * @return Future object which signals the completion of the openat request
 */
[[nodiscard]] inline auto openat(int dirfd, const char *pathname, int flags, mode_t mode = 0)
		-> std::unique_ptr<Future> {
	auto future = std::make_unique<OpenatFuture>(dirfd, pathname, flags, mode);
	future->submit();
	return future;
}

/**
 * @brief synchronous openat mimicking POSIX openat(3)
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param dirfd directory the pathname is interpreted relative to
 * @param pathname path to the file to open
 * @param flags specify access mode
 * @param mode in which the file should be created
 *
 * @return fd on success, -1 on error
 */
inline auto openatAndWait(int dirfd, const char *pathname, int flags, mode_t mode = 0) -> int {
	OpenatFuture future(dirfd, pathname, flags, mode);
	future.submit();
	return future.waitAndSetErrno();
}

/**
 * @brief asynchronous open mimicking POSIX open(3)
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param pathname path to the file to open
 * @param flags specify access mode
 * @param mode in which the file should be created
 *
 * @return Future object which signals the completion of the open request
 */
[[nodiscard]] inline auto open(const char *pathname, int flags, mode_t mode = 0)
		-> std::unique_ptr<Future> {
	auto future = std::make_unique<OpenatFuture>(AT_FDCWD, pathname, flags, mode);
	future->submit();
	return future;
}

/**
 * @brief synchronous open mimicking POSIX open(3)
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param pathname path to the file to open
 * @param flags specify access mode
 * @param mode in which the file should be created
 *
 * @return fd on success, -1 on error
 */
inline auto openAndWait(const char *pathname, int flags, mode_t mode = 0) -> int {
	OpenatFuture future(AT_FDCWD, pathname, flags, mode);
	future.submit();
	return future.waitAndSetErrno();
}

/**
 * @brief Non-blocking close mimicking POSIX close(3)
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param fd file descriptor to close
 *
 * @return Future object which signals the completion of the close request
 */
[[nodiscard]] inline auto close(int fd) -> std::unique_ptr<Future> {
	auto future = std::make_unique<CloseFuture>(fd);
	future->submit();
	return future;
}

/**
 * @brief Blocking close mimicking POSIX close(3)
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param fd file descriptor to close
 *
 * @return 0 on success, -1 on error
 */
inline auto closeAndWait(int fd) -> int {
	CloseFuture future(fd);
	future.submit();
	return future.waitAndSetErrno();
}

/**
 * @brief submit a close request mimicking POSIX close(3) and forget about it
 *
 * This method must be called from inside the emper runtime because it uses
 * the worker-local IoContext
 *
 * @param fd file descriptor to close
 */
inline void closeAndForget(int fd) {
	CloseFuture future(fd);
	future.submitAndForget();
}

#ifdef EMPER_HAS_FS_PATH
/**
 * @brief recursively walk a file tree in parallel
 *
 * This method must be called from inside the emper runtime because it spawn
 * new Fibers.
 *
 * We use c++17 std::filesystem::directory_iterator to read each directory encountered.
 * For each entry in a directory we use @p filter to test if we should call @p fn on
 * the directory entry in a new Fiber.
 * If we encounter a subdirectory we spawn a new fiber recursively walking this subdirectory.
 * The used algorithm is a parallel depth first search.
 *
 * Algorithm
 * ---------
 *
 *```
 * walk(dir, filter, fn)
 *     CPS cps;
 *     walk_helper(cps, dir, filter, fn);
 *     cps.wait();
 *
 * walk_helper(cps, dir, filter, fn)
 *     for(entry: dir)
 *         if (filter(entry))
 *             spawn(fn(entry), cps);
 *         if (entry.is_dir())
 *             spawn(walk_helper(cps, entry, filter, fn), cps);
 *```
 *
 *
 * @param dirpath The start directory where we start the walk
 * @param filter Function to check if a enw Fiber should be spawned executing fn
 *               for a directory entry
 * @param fn Function to execute in a new Fiber with the current directory entry
 */
void recursive_directory_walk(const fs::path &dirpath,
															std::function<bool(const fs::directory_entry &dirent)> filter,
															std::function<void(fs::directory_entry dirent)> fn);
#endif

/**
 * @brief Fiber accepting TCP connections and scheduling handler fibers
 *
 * @param host The host address the socket should be listening on
 * @param port The port the socket should be listening on
 * @param handler A function or lambda the started handler Fibers should execute.
 *        It is called with the file descriptor returned by accept.
 * @param backlog The TCP connection backlog.
 *
 * @return nullptr on error, otherwise the TCP listener Fiber
 */
auto tcp_listener(std::string &host, std::string &port, const std::function<void(int)> &handler,
									int backlog = 1024) -> Fiber *;
}	 // namespace emper::io
