// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#include "io/IoContext.hpp"

#include <liburing.h>						// for io_uring_submit, io_uring_get_sqe
#include <liburing/io_uring.h>	// for io_uring_cqe, io_uring_params, IORI...
#include <sys/eventfd.h>				// for eventfd
#include <unistd.h>							// for close

#include <array>
#include <atomic>		// for atomic, __atomic_base
#include <cassert>	// for assert
#include <cerrno>		// for errno, ECANCELED, EBUSY, EAGAIN, EINTR
#include <cstring>	// for memset
#include <memory>
#include <ostream>	// for basic_osteram::operator<<, operator<<
#include <string>
#include <utility>
#include <vector>

#include "CallerEnvironment.hpp"	// for CallerEnvironment, EMPER, ANYWHERE
#include "Common.hpp"							// for unlikely, DIE_MSG_ERRNO, DIE_MSG
#include "Debug.hpp"							// for LOGD
#include "Emper.hpp"							// for DEBUG, IO_URING_SQPOLL
#include "Fiber.hpp"
#include "Runtime.hpp"
#include "emper-common.h"
#include "io/Future.hpp"	// for Future, operator<<, Future::State
#include "io/GlobalIoContext.hpp"
#include "io/Stats.hpp"	 // for Stats, nanoseconds
#include "lib/TaggedPtr.hpp"
#include "sleep_strategy/PipeSleepStrategy.hpp"

using emper::lib::TaggedPtr;
using emper::sleep_strategy::PipeSleepStrategy;

namespace emper::io {

thread_local IoContext *IoContext::workerIo = nullptr;

auto IoContext::prepareFutureChain(Future &future, unsigned chain_length) -> unsigned {
	unsigned total_chain_length = chain_length;

	if (unlikely(chain_length > *this->ring.sq.kring_entries)) {
		DIE_MSG("Future chain is to long for the io_uring's SQ");
	}

	if (future.dependency) {
		total_chain_length = prepareFutureChain(*future.dependency, chain_length + 1);
	}

	if constexpr (emper::DEBUG) {
		reqs_in_uring++;
		uringFutureSet.insert(&future);
	}

	struct io_uring_sqe *sqe = io_uring_get_sqe(&ring);
	// The kernel consumes all sqes from the SQ during io_uring_enter
	// If we can't get a sqe the chain was to long for our SQ
	assert(sqe);

	future.prepareSqe(sqe);

	// remember the prepared sqe to invalidate it if it was not properly submitted
	preparedSqes.push_back(sqe);

	// we should start a new Fiber executing callback on completion
	if (future.callback) {
		LOGD("prepare " << future << " Callback " << future.callback);
		// set the Callback's affinity to the submitting worker
		if (worker) future.callback->affinity = worker->getWorkerId();
		io_uring_sqe_set_data(sqe,
													TaggedPtr(future.callback, static_cast<uint16_t>(PointerTags::Callback)));

		// Someone wants to be notified about the completion of this Future
	} else if (!future.isForgotten()) {
		io_uring_sqe_set_data(sqe, TaggedPtr(&future, static_cast<uint16_t>(PointerTags::Future)));
	}

	future.state.submitted = true;

	if (future.isDependency()) {
		LOGD("Prepare " << future << " as a dependency");
		sqe->flags |= IOSQE_IO_LINK;
	}

	return total_chain_length;
}

template <CallerEnvironment callerEnvironment>
auto IoContext::submitPreparedSqesAndWait(unsigned wait_nr) -> unsigned {
	// submit the Future to the io_uring
	int submitted = io_uring_submit_and_wait(&ring, wait_nr);

	// Actually I don't know how "unlikely" this is
	if (unlikely(submitted < 0)) {
		if (unlikely(submitted != -EBUSY)) {
			errno = -submitted;
			DIE_MSG_ERRNO("io_uring_submit failed");
		}

		// We can't submit our sqe because the CQ is full
		// in worker thread -> reapCompletions
		// in globalCompleter thread -> TODO: deterministically handle global full CQ
		// for now hope a jam does not happen or will solve itself
		LOGD("io_submit returned EBUSY trying to submit " << io_uring_sq_ready(&ring)
																											<< " in addition to " << reqs_in_uring);

		if constexpr (callerEnvironment == CallerEnvironment::ANYWHERE) {
			// we are not busy looping in the globalIo
			DIE_MSG("Future dropped because global completer SQ is full");
		}

		TIME_NS(
				{
					do {
						reapAndScheduleCompletions();
					} while ((submitted = io_uring_submit(&ring)) == -EBUSY);
				},
				stats.record_io_submit_full_cq);
	}

	return static_cast<unsigned>(submitted);
}

template <CallerEnvironment callerEnvironment>
void IoContext::submitAndWait(Future &future, unsigned wait_nr) {
	LOGD("submitting " << future << (future.dependency ? " and it's dependencies" : ""));
	unsigned prepared = prepareFutureChain(future);

	// submit the Future to the io_uring
	int submitted = submitPreparedSqesAndWait<callerEnvironment>(wait_nr);

	// We submitted some Futures to the io_uring
	// Because we submit every Future right away multiple prepared sqes can only
	// occur for Future chains.
	// If we could not submit the whole chain cancel all non submitted Futures
	// because we can not guaranty the soundness of the chain have to invalidate
	// non submitted but prepared sqes
	// req1 -> invalid_req -> req3
	// will submit only 2 instead of all 3 prepared sqes leaving a sqe for req3
	// in the SQ but unsubmitted
	// See: https://github.com/axboe/liburing/issues/186
	// https://github.com/axboe/liburing/issues/92
	//
	// Since kernel commit cf10960426515 io_uring does no longer submit parts of a
	// broken chains. This means that all sqes in a chain except the broken one will
	// result in cqes with result -ECANCELD and the invalid one will
	// generate a cqe with the appropriate error code
	if (unlikely(static_cast<unsigned>(submitted) < prepared)) {
		unsigned unsubmitted = io_uring_sq_ready(&ring);
		Future *unsubmittedFuture = &future;

		for (unsigned i = 0; i < unsubmitted; ++i) {
			// we prepare sqes for future from the chain end to the head
			// A -> B -> C will result in CQ [C, B, A] and preparedSqes [C, B, A]
			// invalidate/cancel N sqes/Futures from the back
			struct io_uring_sqe *sqe = preparedSqes.back();
			preparedSqes.pop_back();
			// invalidate the prepared sqe
			io_uring_prep_nop(sqe);
			io_uring_sqe_set_data(sqe, nullptr);

			// cancel the according future
			unsubmittedFuture->recordCompletion(stats, -ECANCELED);
			if constexpr (callerEnvironment == EMPER) {
				unsubmittedFuture->complete(-ECANCELED);
			} else {
				unsubmittedFuture->completeFromAnywhere(-ECANCELED);
			}
			unsubmittedFuture = future.dependency;
		}

		// submit all now invalidated sqes
#ifdef NDEBUG
		ATTR_UNUSED
#endif
		unsigned invalidated = submitPreparedSqes<callerEnvironment>();
		assert(invalidated == unsubmitted);
	}

	// we have submitted all our prepared sqes
	preparedSqes.clear();
}

// show the compiler our template incarnations
template void IoContext::submitAndWait<CallerEnvironment::EMPER>(Future &future, unsigned wait_nr);
template void IoContext::submitAndWait<CallerEnvironment::ANYWHERE>(Future &future,
																																		unsigned wait_nr);

template <CallerEnvironment callerEnvironment>
auto IoContext::reapCompletions(ContinuationBuffer &continuationFibers) -> unsigned {
	// Should not be more than the uring_entries count.
	const unsigned CQE_BATCH_COUNT = EMPER_IO_WORKER_URING_ENTRIES;

	unsigned reReapCount = 0;
	uint32_t maxRaceFreeCompleterAttempts = 1;

	using Completion = std::pair<int32_t, void *>;
	// vector to store seen cqes to make the critical section
	// where cq_lock is held as small as possible
	std::array<Completion, CQE_BATCH_COUNT> reapedCompletions;

	// this label is not used for callerEnvironment::ANYWHERE and thus has to be
	// annotated with ATTR_UNUSED
reap_cqes:
	ATTR_UNUSED;

	// never reap completions on the global IoContext
	assert(this != runtime.globalIo);

	LOGD("Reaping completions for worker " << std::to_string(worker->getWorkerId()));
	std::array<struct io_uring_cqe *, CQE_BATCH_COUNT> cqes;

	if constexpr (needsCqLock) {
		// Someone else is currently reaping completions
		if constexpr (callerEnvironment == CallerEnvironment::EMPER) {
			if (unlikely(!cq_lock.try_lock())) {
				LOGD("worker unsuccessful try_lock");
				return 0;
			}
		} else {
			// We can not reap completions of this IoContext to not race
			// with the sleeping worker.
			if (waitInflight.load(std::memory_order_acquire)) {
				LOGD("Not reaping worker " << std::to_string(worker->getWorkerId())
																	 << " since this worker is already waiting for its CQEs");
				return 0;
			}

			if (!cq_lock.try_lock_or_increment()) {
				LOGD("Global completer unsuccessful try_lock_or_increment");
				return 0;
			}
		}
	}

	unsigned count = io_uring_peek_batch_cqe(&ring, cqes.data(), CQE_BATCH_COUNT);

	for (unsigned i = 0; i < count; ++i) {
		struct io_uring_cqe *cqe = cqes[i];
		void *cqe_data = io_uring_cqe_get_data(cqe);

		auto &reapedCompletion = reapedCompletions[i];
		reapedCompletion.first = cqe->res;
		reapedCompletion.second = cqe_data;
	}

	io_uring_cq_advance(&ring, count);

	uint32_t globalCompleterAttempts;
	if constexpr (needsCqLock) {
		globalCompleterAttempts = cq_lock.unlock();
	}

	LOGD("got " << count << " cqes from worker " << worker->getWorkerId() << "'s io_uring");

	if constexpr (emper::DEBUG) {
		assert(count <= reqs_in_uring);
		reqs_in_uring -= count;
	}

	// A naive try lock protecting a worker's IoContext's cq is racy.
	// While a worker is holding the lock additional completions could arrive
	// which the worker does not observe because it could be already finished iterating.
	// In the case that the worker still holds the lock preventing the globalCompleter
	// from reaping the additional completions we have a lost wakeup possibly leading
	// to a completely sleeping runtime with runnable completions in a worker's IoContext.

	// To prevent this race the cq_lock counts the unsuccessful tries from
	// the globalCompleter.
	// If a worker observes that the globalCompleter tried to reapCompletions
	// more than twice we know that a lost wakeup could have occurred and we try to
	// reap again.

	// In the case a lost wakeup was possible we schedule our reaped cqes
	// and try again.

	// On all cq iteration after the first we expect no globalCompleterAttempt
	// or in other words a single globalCompleterAttempt attempt means
	// additional completions arrive and lost wakeup was possible again.

	stats.record_reaps<callerEnvironment>(count);

	unsigned posInBuf = 0;
	for (unsigned i = 0; i < count; ++i) {
		auto &completion = reapedCompletions[i];
		auto res = completion.first;
		auto *cqe_data = completion.second;

		TaggedPtr tptr(cqe_data);
		// Got a CQE for a forgotten Future.
		if (!tptr) {
			continue;
		}

		auto tag = static_cast<PointerTags>(tptr.getTag());
		switch (tag) {
			case PointerTags::NewWorkWsq:
			case PointerTags::NewWorkAq: {
				LOGD("Got new work notification");

				auto *pipeSleepStrategy =
						reinterpret_cast<PipeSleepStrategy *>(&runtime.getWorkerSleepStrategy());
				statsIncr(pipeSleepStrategy->stats->wakeupDueToNotify);

				// Reset flag to indicate that a new sleep cqe must be prepared
				// and allow the completer to reap completions again
				waitInflight.store(false, std::memory_order_release);
				break;
			}

			case PointerTags::Callback: {
				auto *callback = tptr.getPtr<Future::CallbackInternal>();
				LOGD("Create new callback fiber for " << callback);
				auto *callbackFiber = Fiber::from(
						[&c = *callback, res] {
							c(res);
							delete &c;
						},
						&callback->affinity);

				continuationFibers[posInBuf++] = callbackFiber;
			} break;

			case PointerTags::Future: {
				auto *future = tptr.getPtr<Future>();
				// assert that the future was previously in the uringFutureSet
				assert(uringFutureSet.erase(future) > 0);

				future->recordCompletion(stats, res);
				Fiber *continuation = future->completeAndGetContinuation(res);
				if (continuation) {
					continuationFibers[posInBuf++] = continuation;
				}
			} break;

			default:
				DIE_MSG("Unknown pointer tag encountered: " << (int)tag);
				break;
		}
	}

	// check if lost wakeup was possible
	if constexpr (needsCqLock && callerEnvironment == CallerEnvironment::EMPER) {
		bool reReap = false;
		// TODO: How sure are we that this is unlikely?
		if (unlikely(globalCompleterAttempts > maxRaceFreeCompleterAttempts)) {
			// In all CQ iteration after the first we expect no further globalCompleter attempts
			maxRaceFreeCompleterAttempts = 0;
			reReap = true;
		} else if (count == CQE_BATCH_COUNT) {
			// We reaped a full batch, this means there could be potentially
			// more CQEs in the completion queue.
			reReap = true;
		}

		if (reReap) {
			// schedule all already collected continuation fibers
			runtime.schedule(continuationFibers.data(), posInBuf);

			reReapCount++;

			goto reap_cqes;
		}

		stats.record_reReapCount(reReapCount);
	}

	return posInBuf;
}

// Show the compiler our template incarnations this is needed again because
// reapCompletions<CallerEnvironment::ANYWHERE> is now called from GlobalIoContext.cpp
template auto IoContext::reapCompletions<CallerEnvironment::ANYWHERE>(
		ContinuationBuffer &contiunationFibers) -> unsigned;
template auto IoContext::reapCompletions<CallerEnvironment::EMPER>(
		ContinuationBuffer &continuationFibers) -> unsigned;

IoContext::IoContext(Runtime &runtime, size_t uring_entries) : runtime(runtime) {
	struct io_uring_params params;
	memset(&params, 0, sizeof(params));

	if constexpr (emper::IO_URING_SQPOLL) {
		params.flags |= IORING_SETUP_SQPOLL;
	}

	if constexpr (emper::IO_URING_SHARED_WQ) {
		auto *gio = runtime.globalIo;
		if (gio) {
			params.flags |= IORING_SETUP_ATTACH_WQ;
			params.wq_fd = gio->ring.ring_fd;
		}
	}

	auto ret = io_uring_queue_init_params(uring_entries, &ring, &params);
	if (ret < 0) {
		DIE_MSG_ERRNO("io_uring_queue_init failed");
	}
	LOGD("Ring fd is " << ring.ring_fd);

	if (!(params.features & IORING_FEAT_NODROP)) {
		LOGW("kernel das not support IORING_FEAT_NODROP. We may loose IO requests");
	}

	if constexpr (emper::IO_URING_SQPOLL) {
		// included in liburing since 41e0d97cb23667df000ce76789297f4e06134a28
#ifndef IORING_FEAT_SQPOLL_NONFIXED
#define IORING_FEAT_SQPOLL_NONFIXED (1U << 7)
#endif
		if (!(params.features & IORING_FEAT_SQPOLL_NONFIXED)) {
			DIE_MSG("io_uring_sqpoll defined but kernel does not support IORING_FEAT_SQPOLL_NONFIXED");
		}
	}

	// reserve space for a full SQ
	preparedSqes.reserve(*this->ring.sq.kring_entries);

	// This eventfd will be registered to retrieve completion notifications when the Runtime
	// calls registerWorkerIo() on the globalIo
	// Or it will be used to terminate the globalIo
	notificationEventFd = eventfd(0, 0);
	if (unlikely(notificationEventFd < 0)) {
		DIE_MSG_ERRNO("creating eventfd for io_uring failed");
	}
}

IoContext::~IoContext() {
	io_uring_queue_exit(&ring);
	// TODO: check if this is safe
	::close(notificationEventFd);
}
}	 // namespace emper::io
