// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#pragma once

#include <liburing.h>	 // for io_uring

#include <array>
#include <atomic>			 // for atomic
#include <cassert>		 // for assert
#include <cstddef>		 // for size_t
#include <cstdint>		 // for uint64_t
#include <functional>	 // for less
#include <vector>

#include "CallerEnvironment.hpp"	// for CallerEnvironment, EMPER
#include "Common.hpp"
#include "Debug.hpp"	// for LogSubsystem, LogSubsystem::IO, Logger
#include "Emper.hpp"
#include "Runtime.hpp"	// for Runtime
#include "Worker.hpp"
#include "emper-config.h"					// for EMPER_IO_WORKER_URING_ENTRIES
#include "io/Stats.hpp"						// for Stats
#include "lib/adt/LockedSet.hpp"	// for LockedSet

class AbstractWorkStealingScheduler;
class Fiber;

#ifdef EMPER_IO_CQ_LOCK_COUNTING_TRY_LOCK
#include "lib/sync/CountingTryLock.hpp"
using CqLock = emper::lib::sync::CountingTryLock;

#elif defined EMPER_IO_CQ_LOCK_MUTEX
#include <mutex>

#include "lib/sync/PseudoCountingTryLock.hpp"
using CqLock = emper::lib::sync::PseudoCountingTryLock<std::mutex>;

#elif defined EMPER_IO_CQ_LOCK_SPIN_LOCK
#include "lib/sync/PseudoCountingTryLock.hpp"
#include "lib/sync/SpinLock.hpp"
using CqLock = emper::lib::sync::PseudoCountingTryLock<emper::lib::sync::SpinLock>;

#else
#error Uknown cq lock implementation
#endif

namespace emper::sleep_strategy {
class PipeSleepStrategy;
}

namespace emper::io {
class Future;

class IoContext : public Logger<LogSubsystem::IO> {
	friend class ::Runtime;
	friend class ::AbstractWorkStealingScheduler;
	friend class Future;
	friend class SendFuture;
	friend class RecvFuture;
	// We need this friend declaration to access protected members through
	// IoContext pointer in GlobalIoContext::globalCompleterFunc
	friend class GlobalIoContext;

	friend class emper::sleep_strategy::PipeSleepStrategy;

 protected:
	// Remember the Runtime which created the IoContext
	Runtime &runtime;

	static thread_local IoContext *workerIo;
	static constexpr bool needsCqLock =
			emper::IO_COMPLETER_BEHAVIOR != emper::IoCompleterBehavior::none;
	// TryLock protecting the completion queue of ring.
	ALIGN_TO_CACHE_LINE CqLock cq_lock;
	struct io_uring ring;

	// In a worker's IoContext This eventfd is registered with the io_uring to get completion
	// notifications in the globalCompleter.
	// In a GlobalIoContext this eventfd is used to initiate termination
	int notificationEventFd;
	uint64_t notificationEventFdBuf;

	// Remember the worker object for this IoContext
	Worker *worker;

	// Flag to indicate that we already have a sleep related request in the io_uring.
	// Gets set by the worker on WaitdfdSleepStrategy::sleep() and
	// reset when reaping a completion containing a NewWork{Wsq,Aq} TaggedPtr.
	// It also prevents the completer from reaping completions on this IoContext
	// to prevent races between the worker and the completer.
	std::atomic<bool> waitInflight = false;

	Stats stats;

	// Members useful for debugging
	std::atomic<long> reqs_in_uring = 0;

	// We need to keep track of prepared sqes to invalidate them on short submission
	// If a user wants to submit a chain of 5 Futures and the third is invalid
	// io_uring_submit will only submit 3 of 5 prepared sqes leaving the 2 unsubmitted in
	// the SQ.
	// Those unsubmitted sqes would be submitted on the next call of io_uring_submit
	// possibly breaking the guaranties of a chain (cancel all dependent operations).
	// Further more having unsubmitted sqes in the SQ prevents SQ-sized chains from beeing
	// prepared.
	// To prevent all those problem we keep track of the prepared sqes and "invalidate"
	// (prepare as NOP with a nullptr as user_data) submit all unsubmitted  sqes
	// and cancel their futures
	std::vector<struct io_uring_sqe *> preparedSqes;

	// set containing all Futures currently in the io_uring for debugging
	emper::lib::adt::LockedSet<Future *> uringFutureSet;

	/**
	 * @brief recursively prepare a Future and its dependencies
	 *
	 * Futures prepared as dependency will have disabled partial completion
	 * support and are submitted to io_uring with the IOSQE_IO_LINK flag
	 * set.
	 * Abort on preparing a longer list then there a places in the io_uring.
	 *
	 * @param future The future starting the Future chain to prepare
	 * @param chain_length Recursion parameter tracking the resulting chain length
	 *
	 * @return the number of prepared Futures
	 */
	auto prepareFutureChain(Future &future, unsigned chain_length = 1) -> unsigned;

	enum class PointerTags : uint16_t { Future, Callback, NewWorkAq, NewWorkWsq };

	/**
	 * @brief submit prepared sqes possibly reaping completions if CQ is full
	 *
	 * @param wait_nr The number of cqes which should be awaited before returning from
	 *        io_uring_submit_and_wait
	 *
	 * @return the number of submitted Futures
	 */
	template <CallerEnvironment callerEnvironment>
	auto submitPreparedSqesAndWait(unsigned wait_nr) -> unsigned;

	/**
	 * @brief submit prepared sqes possibly reaping completions if CQ is full
	 *
	 * @return the number of submitted Futures
	 */
	template <CallerEnvironment callerEnvironment>
	auto submitPreparedSqes() -> unsigned {
		return submitPreparedSqesAndWait<callerEnvironment>(0);
	}

	/**
	 * @brief Submit a future for asynchronous completion in this IoContext
	 *
	 * This function is not part of the public API of the IoContext because a
	 * user code should not be able block the worker thread in the IoContext.
	 *
	 * @param future The Future which should be completed. The Future object
	 *        must stay valid until it is completed.
	 * @param wait_nr The number of cqes which should be awaited before returning from
	 *        io_uring_submit_and_wait
	 */
	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void submitAndWait(Future &future, unsigned wait_nr);

	inline void setWorkerIo(Worker *worker) {
		// remember our worker
		this->worker = worker;
		workerIo = this;

		if (emper::STATS) {
			stats.workerId = worker->getWorkerId();
		}
	}

 public:
	IoContext(Runtime &runtime, size_t uring_entries);
	IoContext(Runtime &runtime) : IoContext(runtime, EMPER_IO_WORKER_URING_ENTRIES){};
	~IoContext();

	/**
	 * @brief get IoContext of the current worker
	 *
	 * This function must be called from with in the emper runtime
	 *
	 * @return the thread_local IoContext of the current worker
	 */
	[[nodiscard]] static inline auto getWorkerIo() -> IoContext * {
		assert(Runtime::inRuntime());
		return workerIo;
	}

	/**
	 * @brief get IoContext for the CallerEnvironment
	 *
	 * This function is for now not independent of a global Runtime reference
	 * because it uses Runtime::getRuntime() if Runtime::getRuntime() would
	 * use a thread_local variable instead of a global one this function
	 * would be like the rest of IoContext independent of a global Runtime
	 * reference
	 *
	 * @return the either the worker's IoContext or the GlobalIoContext
	 */
	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	[[nodiscard]] static inline auto getIo() -> IoContext * {
		if constexpr (callerEnvironment == CallerEnvironment::EMPER) {
			return IoContext::getWorkerIo();
		} else {
			// we use a reinterpret_ cast here because GlobalIoContext is incomplete
			// at this point but we can't include GlobalIoContext.hpp because it
			// would introduce a cyclic dependency.
			return reinterpret_cast<IoContext *>(Runtime::getRuntime()->globalIo);
		}
	}

	/**
	 * @brief Submit a future for asynchronous completion in this IoContext
	 *
	 * @param future The Future which should be submited to the io_uring.
	 *        The Future object must stay valid until it is completed.
	 */
	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void submit(Future &future) {
		submitAndWait(future, 0);
	}

	/**
	 * @brief Submit a future for asynchronous completion in this IoContext
	 *
	 * @param begin Iterator pointing at the start of the Future collection
	 * @param end Iterator pointing at the end of the Future collection
	 *
	 * TODO: gracefully handle cases where there would be more sqes prepared than
	 * there is room in the SQ
	 */
	template <typename InputIt, CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	auto submit(InputIt begin, InputIt end) -> unsigned {
		for (InputIt cur = begin; cur != end; ++cur) {
			auto future = *cur;
			prepareFutureChain(*future);
		}
		return submitPreparedSqes<callerEnvironment>();
	}

	/**
	 * @brief Submit a future for asynchronous completion in this IoContext and reap completions
	 *
	 * Submit a Future from within the Runtime and reap possible inline completions
	 * directly afterwards
	 *
	 * @param future The Future which should be submited to the io_uring.
	 *        The Future object must stay valid until it is completed.
	 */
	void submitAndReap(Future &future) {
		assert(Runtime::inRuntime());
		submit<CallerEnvironment::EMPER>(future);
		// io_uring will try to synchronously complete any IO request before
		// offloading it to the async backend. See io_uring_enter(2).
		// Immediate offloading can be enforced by setting the IOSQE_ASYNC flag in the sqe.
		// Try to reap a possible synchronous completions.
		reapAndScheduleCompletions();
	}

	static const unsigned CQE_BATCH_COUNT = EMPER_IO_WORKER_URING_ENTRIES;
	using ContinuationBuffer = std::array<Fiber *, CQE_BATCH_COUNT>;

	/**
	 * @brief Collect all fibers waiting on completed IO
	 *
	 * @param[in, out] continuationFibers Buffer big enough to hold all possible continuation Fibers.
	 *                 Passing the buffer fomr the caller to the calle allows the buffer to
	 *                 be statically allocated.
	 *
	 * @return The number of continuation Fibers
	 */
	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	auto reapCompletions(ContinuationBuffer &continuationFibers) -> unsigned;

	/**
	 * @brief Schedule all fibers waiting on completed IO
	 */
	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void reapAndScheduleCompletions() {
		ContinuationBuffer completionBuf;
		unsigned ncompletions = reapCompletions<callerEnvironment>(completionBuf);
		if (ncompletions > 0) {
			auto *fibers = completionBuf.data();

			if constexpr (callerEnvironment == CallerEnvironment::EMPER) {
				runtime.schedule(fibers, ncompletions);
			} else {
				runtime.scheduleFromAnywhere(fibers, ncompletions);
			}
		}
	}

	auto getStats() -> Stats & { return stats; };
};
}	 // namespace emper::io
