// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer, Florian Schmaus
#include "io/Stats.hpp"

#include <initializer_list>	 // for initializer_list
#include <iomanip>					 // for operator<<, setfill, setw
#include <string>						 // for operator<<, to_string
#include <utility>					 // for pair

#include "io/IoContext.hpp"
#include "io/Operation.hpp"	 // for Operation, operator<<, Operation::RECV
#include "lib/math.hpp"

namespace emper::io {

auto operator<<(std::ostream& os, const Stats::CompletionType& t) -> std::ostream& {
	switch (t) {
		case Stats::CompletionType::FullCompletion:
			os << "full completion";
			break;
		case Stats::CompletionType::ErrorCompletion:
			os << "error completion";
			break;
		case Stats::CompletionType::PartialCompletion:
			os << "partial completion";
			break;
		case Stats::CompletionType::PartialResubmission:
			os << "partial resubmission";
			break;
		case Stats::CompletionType::IncrementalCompletion:
			os << "incremental completion";
			break;
		case Stats::CompletionType::IncrementalError:
			os << "incremental error";
			break;
		default:
			abort();
	}
	return os;
}

auto operator<<(std::ostream& os, const Stats& s) -> std::ostream& {
	if (s.workerId == Stats::GLOBAL_COMPLETER_ID) {
		bool futuresToPrint = false;
		std::stringstream ss;
		ss << "# global IO Stats: #" << std::endl;
		// Only PartialCompletableFutures can submitted to the globalIo
		for (const auto& op : {Operation::SEND, Operation::READ, Operation::WRITE}) {
			const auto& op_map = s.io_uring_completions.at(op);

			ss << op << ": ";
			ss << std::left << std::setfill(' ') << std::setw(8) << ss.str();

			for (const auto& tp : op_map) {
				auto completionType = tp.first;
				auto count = tp.second;

				if (count > 0) {
					futuresToPrint = true;
				}

				ss << count << " " << completionType << " ";
			}
			ss << std::endl;
		}

		if (futuresToPrint) {
			os << ss.str();
		}

		return os;
	}

	if (s.workerId == Stats::AVG_ID) {
		os << "# AVG IO worker Stats: #" << std::endl;
	} else {
		os << "# IO Stats of worker " << std::to_string(s.workerId) << ": #" << std::endl;
	}

	for (const auto& pp : s.io_uring_completions) {
		const auto& op = pp.first;
		const auto& op_map = pp.second;

		std::stringstream ss;
		ss << op << ": ";
		os << std::left << std::setfill(' ') << std::setw(16) << ss.str();

		for (const auto& tp : op_map) {
			auto completionType = tp.first;
			auto count = tp.second;
			os << count << " " << completionType << " ";
		}

		os << std::endl;
	}

	if (s.workerId == Stats::GLOBAL_COMPLETER_ID) {
		return os;
	}

	os << "io_uring_submit reap completion loops: " << s.io_submit_full_cq << " taking "
		 << s.io_submit_full_cq_running_mean << "ns on average" << std::endl;

	os << "worker reaped completions " << s.worker_reap << " times reaping "
		 << s.worker_reaped_completions << " completions in total" << std::endl;

	os << "global completer thread reaped completions " << s.completer_reap << " times reaping "
		 << s.completer_reaped_completions << " completions in total" << std::endl;
	os << "re-reap count average [count]: " << s.reReapCount_average << std::endl
		 << "re-reap count max: " << s.reReapCount_max << std::endl
		 << "re-reap last 10:";
	for (unsigned i : s.reReapCount_last) {
		os << " " << std::to_string(i);
	}
	os << std::endl;

	return os;
}

void Stats::printStats(IoContext* globalIoContext,
											 const std::vector<IoContext*>& workerIoContexts) {
	if (globalIoContext) {
		std::cout << globalIoContext->getStats();
	}

	// Use a stats object to calculate the averages
	Stats avgs;
	// mark as averages
	avgs.workerId = Stats::AVG_ID;

	// calculate avgs
	int i = 1;
	for (auto* workerIoContext : workerIoContexts) {
		auto* stats = &workerIoContext->getStats();
		for (const auto& op_pair : stats->io_uring_completions) {
			auto op = op_pair.first;
			auto op_map = op_pair.second;

			for (const auto& compl_pair : op_map) {
				auto compl_typ = compl_pair.first;
				auto compl_count = compl_pair.second;
				math::calcRunningAvg(avgs.io_uring_completions[op][compl_typ], i, compl_count);
			}
		}

		math::calcRunningAvg(avgs.io_submit_full_cq, i, stats->io_submit_full_cq);
		math::calcRunningAvg(avgs.io_submit_full_cq_running_mean, i,
												 stats->io_submit_full_cq_running_mean);

		math::calcRunningAvg(avgs.completer_reap, i, stats->completer_reap);
		math::calcRunningAvg(avgs.completer_reaped_completions, i, stats->completer_reaped_completions);

		math::calcRunningAvg(avgs.worker_reap, i, stats->worker_reap);
		math::calcRunningAvg(avgs.worker_reaped_completions, i, stats->worker_reaped_completions);
		++i;
	}

	// print avgs
	std::cout << avgs << std::endl;
}

}	 // namespace emper::io
