// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus, Florian Fischer
#pragma once

#include <pthread.h>

#include <cstring>
#include <queue>

#include "Common.hpp"

namespace lib::adt {

static void aquire_wrlock(pthread_rwlock_t& lock) {
	int err = pthread_rwlock_wrlock(&lock);
	if (unlikely(err)) {
		DIE_MSG("pthread_rwlock_wrlock failed: " << strerror(err));
	}
}

template <typename I>
class RwLockUnboundedQueue {
 private:
	pthread_rwlock_t lock;

	std::queue<I*> queue;

 public:
	RwLockUnboundedQueue() {
		int err = pthread_rwlock_init(&lock, nullptr);
		if (unlikely(err)) {
			DIE_MSG("pthread_rwlock_init failed: " << strerror(err));
		}
	}

	~RwLockUnboundedQueue() {
		int err = pthread_rwlock_destroy(&lock);
		if (unlikely(err)) {
			DIE_MSG("pthread_rwlock_destroy failed: " << strerror(err));
		}
	}

	void enqueue(I* item) {
		aquire_wrlock(lock);
		queue.push(item);
		pthread_rwlock_unlock(&lock);
	}

	template <class InputIt>
	void insert(InputIt begin, InputIt end) {
		aquire_wrlock(lock);
		for (; begin != end; ++begin) {
			queue.push(*begin);
		}
		pthread_rwlock_unlock(&lock);
	}

	void insert(I** items, unsigned count) {
		aquire_wrlock(lock);
		for (unsigned i = 0; i < count; ++i) {
			queue.push(items[i]);
		}
		pthread_rwlock_unlock(&lock);
	}

	auto dequeue() -> I* {
		I* res = nullptr;

		int err = pthread_rwlock_rdlock(&lock);
		if (unlikely(err)) {
			DIE_MSG("pthread_rwlock_rdlock failed: " << strerror(err));
		}

		if (queue.empty()) {
			goto unlock_and_return;
		}

		// try to upgrade to wrlock
		err = pthread_rwlock_trywrlock(&lock);
		if (err) {
			if (unlikely(err != EBUSY)) {
				DIE_MSG("pthread_rwlock_trylock failed: " << strerror(err));
			}

			// drop the read lock and aquire a write lock
			aquire_wrlock(lock);

			if (queue.empty()) {
				goto unlock_and_return;
			}
		}

		// we certainly hold the wrlock here
		res = queue.front();
		queue.pop();

	unlock_and_return:
		pthread_rwlock_unlock(&lock);
		return res;
	}

	template <class OutputIt>
	auto dequeue(OutputIt begin, OutputIt end) -> size_t {
		int err = pthread_rwlock_rdlock(&lock);
		if (unlikely(err)) {
			DIE_MSG("pthread_rwlock_rdlock failed: " << strerror(err));
		}

		OutputIt cur = begin;
		if (queue.empty()) {
			goto unlock;
		}

		// try to upgrade to wrlock
		err = pthread_rwlock_trywrlock(&lock);
		if (err) {
			if (unlikely(err != EBUSY)) {
				DIE_MSG("pthread_rwlock_trylock failed: " << strerror(err));
			}

			// drop the read lock and aquire a write lock
			aquire_wrlock(lock);

			if (queue.empty()) {
				goto unlock;
			}
		}

		for (; !queue.empty() && cur != end; ++cur) {
			*cur = queue.front();
			queue.pop();
		}

	unlock:
		pthread_rwlock_unlock(&lock);
		return cur - begin;
	}
};
}	 // namespace lib::adt
