// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#pragma once

#include <linux/futex.h>
#include <sys/syscall.h>
#include <unistd.h>

#include <atomic>

#include "Common.hpp"

namespace emper::lib::sync {

// ATTENTION: This semaphore implementation is prone to spurious wakeups!
// SpuriousFutexSemaphore::wait() may return spurious without a call to
// SpuriousFutexSemaphore::notify().
// Use this class only where it is safe to have spurious wakeups.
// It could be used to wakeup and suspend worker threads and does so maximal efficient
// by only using a single atomic operation and a possible system call per method.
class SpuriousFutexSemaphore {
 public:
	using CounterType = long;

 private:
	// >= 0 means no waiters
	// < 0 means there are waiters
	std::atomic<CounterType> counter;

 public:
	SpuriousFutexSemaphore(CounterType counter = 0) : counter(counter) {}

	[[nodiscard]] inline auto getValue() const -> CounterType {
		return counter.load(std::memory_order_relaxed);
	}

	inline void notify_many(CounterType count) {
		// Using std::memory_order_relaxed is fine since we don't
		// give any memory synchronization guaranties because of the
		// possible spurious wakeups
		CounterType c = counter.fetch_add(count, std::memory_order_relaxed);
		if (c < 0) {
			syscall(SYS_futex, &counter, FUTEX_WAKE, count, 0, 0, 0);
		}
	}

	inline void notify() { notify_many(1); }

	inline void wait() {
		// See fetch_add comment.
		CounterType c = counter.fetch_sub(1, std::memory_order_relaxed) - 1;
		while (c < 0) {
			long err = syscall(SYS_futex, &counter, FUTEX_WAIT, c, 0, 0, 0);
			// The futex(2) manpage says that applications should check the futex word
			// again because it is possible to have spurious wakeups.
			// We can't check the futex value because there is no way to decided
			// if someone called notify.
			// A sound generic Semaphore needs to split its atomic counter into
			// a semaphore value and a waiter count part.
			if (!err) {
				return;
			}
			c = getValue();
		}
	}
};

}	 // namespace emper::lib::sync
