// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Schmaus
#include "AbstractWorkStealingScheduler.hpp"

#include <algorithm>
#include <array>
#include <cassert>
#include <ostream>	// for operator<<, basic_ostream<>::__ostream_type

#include "Common.hpp"	 // for unlikely, likely
#include "Debug.hpp"	 // for ABORT
#include "Emper.hpp"	 // for OVERFLOW_QUEUE
#include "Fiber.hpp"
#include "NextFiberResult.hpp"
#include "Runtime.hpp"	// for Runtime
#include "Worker.hpp"
#include "emper-common.h"	 // for workerid_t
#include "io/IoContext.hpp"
#include "lib/TaggedPtr.hpp"
#include "strategies/AbstractWorkStealingStrategy.hpp"
#include "strategies/AbstractWorkStealingWorkerStats.hpp"

using awss = AbstractWorkStealingStrategy;

using emper::io::IoContext;

thread_local AbstractWorkStealingScheduler::WsQueue<AbstractWorkStealingScheduler::QUEUE_SIZE>
		AbstractWorkStealingScheduler::queue;

AbstractWorkStealingScheduler::AbstractWorkStealingScheduler(Runtime& runtime,
																														 RuntimeStrategy& strategy)
		: Scheduler(runtime, strategy) {
	const workerid_t workerCount = runtime.getWorkerCount();
	queues = new AbstractWorkStealingScheduler::WsQueue<QUEUE_SIZE>*[workerCount];

	auto newWorkerHook = [this](workerid_t workerId) { queues[workerId] = &queue; };
	addNewWorkerHook(newWorkerHook);
}

void AbstractWorkStealingScheduler::scheduleViaWorkStealing(Fiber& fiber) {
	bool pushed = queue.pushBottom(&fiber);
	if (unlikely(!pushed)) {
		if constexpr (emper::OVERFLOW_QUEUE) {
			enqueueInAnywhereQueue(fiber);

			emper::statsIncr(awss::stats.scheduledFibersToOverflowQueue);
		} else {
			ABORT("Could not push fiber " << &fiber << " into queue");
		}
	} else if constexpr (emper::STATS) {
		uint64_t queueLength = queue.usedSlots();
		awss::stats.recordScheduledToLocalAndQueueLength(queueLength);
	}

	// Classes using this method are supposed to always invoke this
	// method. Hence we call onNewWork() here.
	onNewWork();
}

auto AbstractWorkStealingScheduler::maybeRecycle(Fiber* fiber) -> bool {
	if (fiber->isRunnable()) return false;

	recycle(fiber);
	return true;
}

auto AbstractWorkStealingScheduler::nextFiberViaAnywhereQueue()
		-> std::optional<std::pair<Fiber*, FiberSource>> {
	const size_t KEEP_FREE_SLOTS = 64;
	const size_t DEQUEUE_FROM_ANYWHERE_MAX = 512;

	const uintptr_t remainingCapacityLocalWsQueue = queue.freeSlots();
	size_t maxFiberCount;
	if (remainingCapacityLocalWsQueue > KEEP_FREE_SLOTS) {
		maxFiberCount = remainingCapacityLocalWsQueue - KEEP_FREE_SLOTS;
		maxFiberCount = std::min(maxFiberCount, DEQUEUE_FROM_ANYWHERE_MAX);
	} else {
		maxFiberCount = 1;
	}

	std::array<Fiber*, DEQUEUE_FROM_ANYWHERE_MAX> fibers;

	size_t fiberCount =
			dequeueFibersFromAnywhereQueue(fibers.begin(), fibers.begin() + maxFiberCount);

	Fiber* res = nullptr;
	size_t fibersLiftedFromAnywhere = 0;
	for (size_t i = 0; i < fiberCount; i++) {
		Fiber* fiber = fibers[i];

		if (maybeRecycle(fiber)) continue;

		// We found a runnable fiber. First check if we already have one
		// fiber saved to be returned.
		if (!res) {
			res = fiber;
			continue;
		}

		// We already have a fiber saved which will be returned. Hence we
		// push the remaining runnable fibers to our work-stealing queue,
		// so that they can be stolen by others.
		[[maybe_unused]] bool pushed = queue.pushBottom(fiber);

		// We have ensured that there is capacity for this fiber, hence we
		// must be able to push it here.
		assert(pushed);

		fibersLiftedFromAnywhere++;
	}

	if (fibersLiftedFromAnywhere) {
		onNewWork();

		if constexpr (emper::STATS) {
			awss::stats.recordFibersLiftedFromAnywhereQueue(fibersLiftedFromAnywhere);
		}
	}

	if (res) return std::make_pair(res, FiberSource::anywhereQueue);

	return std::nullopt;
}

auto AbstractWorkStealingScheduler::tryStealFiberFrom(workerid_t victim)
		-> std::optional<std::pair<Fiber*, FiberSource>> {
	Fiber* fiber;
popTop:
	bool poped = queues[victim]->popTop(&fiber);
	if (poped) {
		emper::statsIncr(awss::stats.nextFiberStolen);

		if (maybeRecycle(fiber)) goto popTop;

		return std::make_pair(fiber, FiberSource::stolen);
	}

	return std::nullopt;
}

auto AbstractWorkStealingScheduler::nextFiberViaWorkStealing() -> std::pair<Fiber*, FiberSource> {
	FiberSource fiberSource = FiberSource::local;
	Fiber* fiber;

popBottom:
	bool poped = queue.popBottom(&fiber);
	if (likely(poped)) {
		emper::statsIncr(awss::stats.nextFiberFromLocal);

		if (maybeRecycle(fiber)) goto popBottom;

		return std::make_pair(fiber, fiberSource);
	}

	auto* const currentWorker = Worker::getCurrentWorker();

	// Try dispatch hint possibly set by the sleep strategy
	{
		if (currentWorker->dispatchHint) {
			const TaggedPtr dispatchHint = currentWorker->dispatchHint;
			currentWorker->dispatchHint = nullptr;
			const auto tag = static_cast<IoContext::PointerTags>(dispatchHint.getTag());

			switch (tag) {
				case IoContext::PointerTags::NewWorkWsq: {
					const auto victim = static_cast<workerid_t>(dispatchHint.getPtrValue());
					const auto stolen = tryStealFiberFrom(victim);
					if (stolen) {
						fiber = (*stolen).first;
						fiberSource = FiberSource::hintWsq;
						goto out;
					}
				} break;

				case IoContext::PointerTags::NewWorkAq: {
					const auto fromAnywhere = nextFiberViaAnywhereQueue();
					if (fromAnywhere) {
						fiber = (*fromAnywhere).first;
						fiberSource = FiberSource::hintAq;
						goto out;
					}
					break;
				}
				default:
					DIE_MSG("invalid dispatch hint");
			}
		}
	}

	// Go into work stealing
	{
		// TODO: Determine if there is a better value than 1/3.
		const float CHECK_ANYWHERE_QUEUE_AT_PERCENTAGE = 0.33;
		const workerid_t myWorkerId = currentWorker->getWorkerId();
		const workerid_t workerCount = runtime.getWorkerCount();
		// NOLINTNEXTLINE(bugprone-narrowing-conversions)
		const workerid_t checkAnywhereQueueAt = workerCount * CHECK_ANYWHERE_QUEUE_AT_PERCENTAGE;

		workerid_t startWorkerId = currentWorker->nextRandomWorkerId();
		// TODO: See how reducing the loop bound affects things.
		for (workerid_t i = 0; i < workerCount; ++i) {
			workerid_t victim = (startWorkerId + i) % workerCount;

			// Don't steal from ourselves.
			if (unlikely(victim == myWorkerId)) continue;

			auto stolenFiber = tryStealFiberFrom(victim);
			if (stolenFiber) return *stolenFiber;

			if constexpr (!emper::CHECK_ANYWHERE_QUEUE_WHILE_STEALING) continue;

			// If we failed to steal from a certain number of victims, check
			// the anywhere queue for new fibers.
			if (i == checkAnywhereQueueAt) {
				auto anywhereQueueFiber = nextFiberViaAnywhereQueue();
				if (anywhereQueueFiber) return *anywhereQueueFiber;
			}
		}
	}

	// Try the "scheduled from anywhere" queue to get work as last resort.
	{
		auto anywhereQueueFiber = nextFiberViaAnywhereQueue();
		if (anywhereQueueFiber) return *anywhereQueueFiber;
	}

	// We where not able to dequeue any fiber if we reach this point.
	fiber = nullptr;

out:
	return std::make_pair(fiber, fiberSource);
}

auto AbstractWorkStealingScheduler::nextFiberResultViaWorkStealing() -> NextFiberResult {
	std::pair<Fiber*, FiberSource> nextFiberWsResult = nextFiberViaWorkStealing();
	return NextFiberResult{
			nextFiberWsResult.first,
			static_cast<uintptr_t>(nextFiberWsResult.second),
	};
}
